from math import inf
from collections import deque, defaultdict
from collections.abc import Mapping
from typing import Generator, TypeVar, Optional
from numbers import Number

import lightning.pytorch as pl
from lightning.pytorch import cli
from optuna.trial import Trial
from optuna.exceptions import TrialPruned

from lightning.pytorch.callbacks import Callback
import lightning as L
import torch
from torch.utils.data import (
    DataLoader,
    Dataset,
    random_split,
    ConcatDataset,
    Subset,
)

from discotime.models import LitSurvModule
from discotime.datasets.utils import SurvDataset

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)


def update_mapping(d: Mapping, u: Mapping) -> Mapping:
    """Update nested mappings recursively.

    Args:
        d (Mapping): dictionary/mapping of things.
        u (Mapping): other dictionary/mapping of things to update `d` with.

    Example:
        >>> foo = {"dog" : {"color" : "black", "age" : 10}}
        >>> bar = {"dog" : {"age" : 11}, "cat" : None}
        >>> recursive_update(foo, bar)
        {'dog': {'color': 'black', 'age': 11}, 'cat': None}
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = update_mapping(d.get(k, {}), v)  # type: ignore
        else:
            d[k] = v  # type: ignore
    return d


def recursive_defaultdict():
    return defaultdict(recursive_defaultdict)


class BaseFabricTrainer:
    """Bare-bones trainer class using Fabric.

    This trainer class supports only a very limited set of the full lightning
    functionality. It can be useful for quick prototyping or for k-fold
    cross-validation as exemplified in the scripts that can be found in the
    ``/experiments`` folder on github.
    """

    def __init__(
        self,
        module: LitSurvModule,
        dset_train: SurvDataset,
        dset_valid: SurvDataset,
        batch_size: int,
    ) -> None:
        self.fabric = L.Fabric(accelerator="cpu", devices=1)
        self.train_dl = DataLoader(
            dset_train, shuffle=True, batch_size=batch_size, drop_last=True
        )
        self.valid_dl = DataLoader(
            dset_valid, shuffle=True, batch_size=batch_size
        )
        self.module = module

    def __iter__(self) -> Generator[torch.Tensor, None, None]:
        self.fabric.launch()
        train_dl = self.fabric.setup_dataloaders(self.train_dl)
        valid_dl = self.fabric.setup_dataloaders(self.valid_dl)
        module, optimizer = self.fabric.setup(
            self.module,
            self.module.configure_optimizers(),
        )

        while True:
            self.module.train()
            with torch.set_grad_enabled(True):
                for batch_idx, batch in enumerate(train_dl):
                    loss = module.training_step(batch, batch_idx)
                    optimizer.zero_grad()
                    self.fabric.backward(loss)
                    optimizer.step()

            with torch.set_grad_enabled(False):
                module.eval()
                losses = (
                    module.validation_step(batch, idx) * len(batch[1])
                    for idx, batch in enumerate(valid_dl)
                )
                yield sum(losses) / len(valid_dl.sampler)


def split_dataset(
    dataset: Dataset[T], k: int = 5
) -> Generator[tuple[ConcatDataset[T], Subset[T]], None, None]:
    """Split dataset into k train/validation splits"""
    splits = deque(random_split(dataset, lengths=[1 / k] * k))
    for _ in range(k):
        dset_valid: Subset[T] = splits.popleft()
        dset_train: ConcatDataset[T] = ConcatDataset(splits)
        yield (dset_train, dset_valid)
        splits.append(dset_valid)


class EarlyStopping:
    """Lightning-free implementation of early stopping.

    To be used with the bare-bones simple trainer that have no support for
    lightning callbacks.
    """

    def __init__(
        self, patience: int, min_delta: float, direction: str = "minimize"
    ) -> None:
        self.patience = patience
        self.min_delta = min_delta
        self.wait = 0

        if direction not in (allowed := {"minimize", "maximize"}):
            raise ValueError(f"direction can be {allowed}, got {direction}")

        self.direction = direction
        self.best = inf if direction == "minimize" else -inf

    def should_stop(self, metric) -> bool:
        match self.direction:
            case "minimize":
                improvement = metric < (self.best - self.min_delta)
            case "maximize":
                improvement = metric > (self.best + self.min_delta)

        if improvement:
            self.wait = 0
            self.best = metric
        else:
            self.wait += 1

        return self.wait >= self.patience


def get_last_intermediate_value(trial: Trial):
    values = trial.storage.get_trial(trial._trial_id).intermediate_values
    return max(values.items(), key=lambda x: x[0])[1]


def clamp(value: float, minvalue: float, maxvalue: float):
    return max(minvalue, min(value, maxvalue))


class OptunaCallback(Callback):
    """Lightning callback for hyperparameter optimization with Optuna.

    Args:
        trial:
            A :class:`~optuna.trial.Trial` corresponding to the current
            evaluation of the objective function.
        monitor:
            An evaluation metric for pruning, e.g., ``val_loss`` or
            ``val_acc``.
    """

    def __init__(
        self,
        trial: Trial,
        monitor: str,
        minvalue: float = -inf,
        maxvalue: float = inf,
    ) -> None:
        super().__init__()

        self.minvalue = minvalue
        self.maxvalue = maxvalue
        self.trial = trial
        self.monitor = monitor

    def on_validation_end(
        self, trainer: pl.Trainer, module: pl.LightningModule
    ) -> None:
        if trainer.sanity_checking:
            return

        metric = trainer.callback_metrics.get(self.monitor)
        if metric is None:
            raise ValueError(
                f"The metric '{self.monitor}' is not in the evaluation logs."
                f"Available metrics are: {trainer.callback_metrics}"
            )

        epoch = module.current_epoch
        self.trial.report(
            clamp(metric.item(), self.minvalue, self.maxvalue),
            step=epoch,
        )

        if not self.trial.should_prune():
            return

        raise TrialPruned(f"Trial was pruned at epoch {epoch}.")



from typing import Optional
from collections.abc import Iterable

import numpy as np
import numpy.typing as npt
import torch
from einops import repeat

Int = int | np.int_
Num = Int | float | np.float_


def _tabulate(a, v, side="left"):
    """Count interval occurences.

    As an example, if side="left" and a=[1, 2, 3], then intervals are
    (-inf, 1), [1, 2), [2, 3) and [3, inf)

    Example:
        >>> a, v = np.array([1, 2, 3]), np.array([0.5, 1, 2.5, 4])
        >>> _tabulate(a, v)
        array([2, 0, 1, 1])
    """
    return np.bincount(np.searchsorted(a, v, side=side))


class KaplanMeier:
    """Simple implementation of the Kaplan-Meier estimator.

    Args:
        time: event times.
        event: event indicator (0/1) where 0 is censoring.

    Example:
        >>> km = KaplanMeier(time=[0, 1.5, 1.3, 3], event=[0, 1, 0, 0])
        >>> km(0)
        array([1.])
        >>> km([0, 1.0, 1.1, 1.5])
        array([1. , 1. , 1. , 0.5])
    """

    def __init__(self, time: Iterable[Num], event: Iterable[Int]) -> None:
        _time, _event = map(np.asarray, (time, event))
        order = np.argsort(_time)
        _time, _event = _time[order], _event[order]

        # observed event times
        tj = np.unique(np.pad(_time[_event == 1], (1, 0)))

        def _pad(a):
            return np.pad(a, (0, tj.size - a.size))

        mj = _pad(_tabulate(tj, _time[_event == 1], side="right")[1:])
        qj = _pad(_tabulate(tj, _time[_event == 0], side="right")[1:])

        nj = np.roll(_time.size - np.cumsum(mj + qj), 1)
        nj[0] = _time.size

        self._sj = np.cumprod((nj - mj) / nj)
        self._nj, self._tj, self._mj = nj, tj, mj

    def __call__(self, time: Num | Iterable[Num]) -> npt.NDArray[np.float_]:
        """Obtain Kaplan-Meier estimates for each timepoint.

        Args:
            time (num | seq[num]): `t`'s for which `km(t)` will be returned.
        """
        _time = np.atleast_1d(np.asanyarray(time))
        return self._sj[np.searchsorted(self._tj, _time, side="right") - 1]

    def percentile(
        self, p: Iterable[Num], dtype=np.float64
    ) -> npt.NDArray[np.float_]:
        """Obtain approximate timepoint t such that P(t) = p.

        The stepwise Kaplan-Meier estimator is piecewise linearly interpolated
        such that unique timepoints can be obtained.
        """
        p = np.atleast_1d(np.asanyarray(p, dtype=dtype))
        if not np.all((0 <= p) & (p <= 1)):
            raise ValueError(
                "p is a probability and should be between 0 and 1"
            )
        return np.interp(1 - p, 1 - self._sj, self._tj, left=0)


class AalenJohansen:
    """Obtain cumulative incidence curves with the Aalen-Johansen method.

    Args:
        time: event times
        event: event indicator (0/1/../c) with 0=censored
        n_causes: how many causes should be included? If None (the default)
            then all observed causes are include.
    """

    def __init__(
        self,
        time: npt.ArrayLike,
        event: npt.ArrayLike,
        n_causes: Optional[int] = None,
    ) -> None:
        time, event = map(np.asarray, (time, event))

        if time.ndim != 1:
            raise ValueError(f"`time` is a {time.ndim}D array.")

        if event.ndim != 1:
            raise ValueError(f"`event` is a {event.ndim}D array.")

        order = np.argsort(time)
        time, event = time[order], event[order]

        # find unique event times and add t=0 if it isn't present
        tj = np.unique(np.pad(time, (1, 0)))

        # number at risk at the start of each interval
        nj = time.size - np.cumsum(_tabulate(tj, time, side="right"))
        nj = nj[:-1]  # drop last

        def _pad(a):
            """Add zeros to end of array to ensure it has same size as `tj`"""
            assert (tj.size - a.size) >= 0
            return np.pad(a, (0, tj.size - a.size), constant_values=0)

        # number lost in each interval to any cause
        mj = _pad(_tabulate(tj, time[event != 0]))

        # lagged survival which starts with P(0) = 1
        sj = np.cumprod((nj - mj) / nj)[:-1]
        sj = np.pad(sj, (1, 0), constant_values=1)

        # cause-specific incidences
        n_risks = n_causes if n_causes else np.max(event)
        ci = np.zeros((tj.size, n_risks))
        for e in range(1, n_risks):
            mcj = _pad(_tabulate(tj, time[event == e]))
            ci[:, e - 1] = np.cumsum(sj * (mcj / nj))

        self._tj = tj
        self._sj = sj
        self._ci = ci

    def __call__(self, timepoints: npt.ArrayLike) -> npt.NDArray[np.float64]:
        """Return cause-specific cumulative incidence at given timepoints."""
        tau = np.asanyarray(timepoints).reshape(-1)
        idx = np.searchsorted(self._tj, tau, side="right")
        idx = np.clip(idx - 1, a_min=0, a_max=(self._tj.size - 1))
        return self._ci[idx]


def interpolate2d(x: torch.Tensor, xp: torch.Tensor, yp: torch.Tensor):
    """Perform stepwise linear interpolation of a discrete function.

    _xp_ and _yp_ are tensors of values used to approximate f: y = f(x). This
    functions uses interpolation to find the value of new points x.

    Args:
        x (:obj:`torch.Tensor`): an 1D tensor real values.
        xp (:obj:`torch.Tensor`): an 1D tensor of real values.
        yp (:obj:`torch.Tensor`): an ND tensor of real values. The length of yp
            along the second axis (dim=1) must have the same length as xp.
    """
    x, xp = x.to(yp), xp.to(yp)  # move to same device
    m = torch.diff(yp, dim=1) / repeat(torch.diff(xp, dim=0), "t -> 1 t 1")
    b = yp[:, :-1, :] - torch.einsum("btr,t->btr", m, xp[:-1])
    idx = torch.clamp_max(torch.searchsorted(xp, x), m.shape[1] - 1)
    return m[:, idx, :] * repeat(x, "t -> 1 t 1") + b[:, idx, :]



from typing import Optional, Any, Iterable, Callable, Dict
from dataclasses import dataclass
import warnings

import torch
from torch.nn import functional as F
from torch import nn
from lightning import pytorch as pl
import numpy as np

from discotime.models.components import Net, negative_log_likelihood
from discotime.metrics import IPA
from discotime.datasets import LitSurvDataModule
from discotime.utils import interpolate2d


@dataclass(kw_only=True)
class ModelConfig:
    learning_rate: float = 1e-3
    activation_function: str = "SiLU"
    """Name of activation function (str). The name needs to match one of the
    activation functions in the ``torch.nn`` module.
    """

    n_sequential_blocks: int = 2
    """Number of neural-network layer blocks.

    Each block is modelled after a ResNet skip-block and contains the following
    key elements: 
    
    .. code-block ::

        Sequential(
            (0): LazyBatchNorm1d()
            (1): LazyLinear()
            (2): DropOut()
            (3): SiLU()
            (4): LazyLinear()
            (5): DropOut()
        )

    Per default the output of the model is ``self.act_fn(x + self.net(x))``
    but the skip part can be removed by setting :attr:`use_skip_connections` to 
    ``False``. :attr:`n_hidden_units` controls the size of the linear layers in 
    each block. See :class:`.components.Block` for more details 
    on the actual implementation.
    """
    n_hidden_units: int = 20
    """Number of neurons in each of the hidden layers.
    
    For ease of of use, the size of each hidden layer have been constrained to
    to be the exact same size as all the others. Default is 20.
    """
    dropout_rate: Optional[float] = None
    """Should ``nn.Dropout()`` be used in each block, and if so what is the
    rate of dropout? If ``None`` then dropout is not used. Default is None.
    """
    batch_normalization: bool = True
    """Use batch normalization in each block? Default is True."""
    use_skip_connections: bool = False
    """Toggle the use of skip-connections in the model blocks."""

    evaluation_grid_size: int = 50
    """Either the length of (int) or the specific grid (seq of floats) at which
    model metrics are calculated. If an integer `n` is passed, then `n` evenly
    distributed timepoint are chosen from the range of the data.
    """
    evaluation_grid_cuts: Optional[int] = None
    """Either the length of (int) or the specific grid (seq of floats) at which
    model metrics are calculated. If an integer `n` is passed, then `n` evenly
    distributed timepoint are chosen from the range of the data.
    """
    n_time_bins: Optional[int] = None
    """Number of time bins in discretization grid.

    If ``None``, which is the default, then we try to get the information from
    the attached datamodule during setup.
    """
    n_risks: Optional[int] = None
    """How many competing risks are we dealing with?

    If ``None``, which is the default, then we try to get the information from
    the attached datamodule during setup.
    """


class LitSurvModule(pl.LightningModule):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__()
        self.config = config
        self.learning_rate = config.learning_rate

        self._loss = negative_log_likelihood
        self._metrics: dict[str, Callable] = {}
        self._eval_grid: Optional[torch.Tensor] = None
        self._data_cuts: Optional[torch.Tensor] = None
        self._model: Optional[nn.Module] = None
        self._datamodule: Optional[LitSurvDataModule] = None

        self._test_step_outputs: list[tuple[torch.Tensor, ...]] = []
        self._validation_step_outputs: list[tuple[torch.Tensor, ...]] = []

        self.save_hyperparameters()

    @property
    def datamodule(self) -> LitSurvDataModule:
        if self._datamodule is not None:
            return self._datamodule
        return self.trainer.datamodule  # type: ignore

    @datamodule.setter
    def datamodule(self, value: LitSurvDataModule) -> None:
        self._datamodule = value

    @property
    def model(self) -> nn.Module:
        if self._model is None:
            raise AssertionError(
                "`self.model` have not been instantiated."
                "Try calling `self.setup()` first."
            )
        return self._model

    @property
    def eval_grid(self) -> torch.Tensor:
        if self._eval_grid is None:
            raise AssertionError(
                (
                    "`self.eval_grid` has not been instantiated."
                    "Try calling `self.setup()` first."
                )
            )
        return self._eval_grid

    @property
    def data_cuts(self) -> torch.Tensor:
        if self._data_cuts is None:
            raise AssertionError(
                (
                    "`self.data_cuts` has not been instantiated."
                    "Try calling `self.setup()` first."
                )
            )
        return self._data_cuts

    def setup(self, stage):
        """Called at the beginning of fit (train + validate), validate, test,
        or predict. This is where the PyTorch model gets instantiated.

        Args:
            stage (`str`): either ``"fit"``, ``"validate"`` ``"test"`` or
                ``"predict"``
        """
        if not isinstance((conf := self.config), ModelConfig):
            raise TypeError("`config` has wrong type!")

        def _get_data_attribute(attr: str) -> int:
            """Get attribute from config or attached datamodule."""
            if (value := getattr(conf, attr)) is None:
                try:
                    self.datamodule.setup()  # fit transformers
                    return getattr(self.datamodule, attr)
                except:
                    raise AssertionError(f"`{attr}` could not be found.")
            return value

        self.n_time_bins = _get_data_attribute("n_time_bins")
        self.n_risks = _get_data_attribute("n_risks")

        if not self._model:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._model = Net(
                    n_out_features=self.n_time_bins * (self.n_risks + 1),
                    n_hidden_units=conf.n_hidden_units,
                    add_skip_connection=conf.use_skip_connections,
                    activation_function=getattr(nn, conf.activation_function),
                    batch_normalization=conf.batch_normalization,
                    dropout_rate=conf.dropout_rate,
                    n_blocks=conf.n_sequential_blocks,
                )

        if stage in {"fit", "test", "validate"}:
            self.datamodule.setup()

            # get attributes from the datamodule
            self._data_cuts = torch.as_tensor(self.datamodule.cuts)
            if conf.evaluation_grid_cuts is not None:
                self._eval_grid = torch.as_tensor(conf.evaluation_grid_cuts)
            else:
                start, end = self.datamodule.time_range
                self._eval_grid = torch.linspace(
                    start=start, end=end, steps=conf.evaluation_grid_size  # type: ignore
                )

            # instantiate metrics
            self._metrics["IPA"] = IPA(
                np.asarray(self.eval_grid, dtype=np.float_), integrate=True
            )

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        checkpoint["model"] = self._model

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        self._model = checkpoint["model"]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x).view(-1, self.n_time_bins, self.n_risks + 1)

    def training_step(
        self, batch: tuple[torch.Tensor, ...], batch_idx: int
    ) -> torch.Tensor:
        x, dt, de, _, _ = batch
        train_loss = self._loss(self(x), dt, de)
        self.log("train_loss", train_loss)
        return train_loss

    def validation_step(
        self, batch: tuple[torch.Tensor, ...], batch_idx: int
    ) -> torch.Tensor:
        x, dt, de, ct, ce = batch
        val_loss = self._loss(self(x), dt, de)
        self.log("val_loss", val_loss)
        # aggregate estimates + labels for batch metrics
        self._validation_step_outputs.append(
            (self._predict_estimates(x, self.eval_grid), ct, ce)
        )
        return val_loss

    def test_step(
        self, batch: tuple[torch.Tensor, ...], batch_idx: int
    ) -> torch.Tensor:
        x, dt, de, ct, ce = batch
        test_loss = self._loss(self(x), dt, de)
        self.log("test_loss", test_loss)
        # aggregate estimates + labels for batch metrics
        self._test_step_outputs.append(
            (self._predict_estimates(x, self.eval_grid), ct, ce)
        )
        return test_loss

    def on_test_epoch_end(self) -> None:
        estimates, ct, ce = map(
            lambda x: torch.cat(x).cpu(), zip(*self._test_step_outputs)
        )
        self._test_step_outputs.clear()  # empty list
        for metric, metric_fn in self._metrics.items():
            values = metric_fn(estimates, ct, ce)
            for cause, value in enumerate(values, start=1):
                self.log(f"test_{metric}_cause{cause}", value)

    def on_validation_epoch_end(self) -> None:
        estimates, ct, ce = map(
            lambda x: torch.cat(x).cpu(), zip(*self._validation_step_outputs)
        )
        self._validation_step_outputs.clear()  # empty list
        for metric, metric_fn in self._metrics.items():
            values = metric_fn(estimates, ct, ce)
            for cause, value in enumerate(values, start=1):
                self.log(f"val_{metric}_cause{cause}", value)

    def _predict_estimates(
        self, x: torch.Tensor, timepoints: torch.Tensor
    ) -> torch.Tensor:
        c_haz = F.softmax(self(x), dim=-1)

        s = torch.cumprod(c_haz[..., [0]], dim=1)
        s_lag = torch.roll(s, shifts=1, dims=1)
        s_lag[:, 0, :] = 1

        # convert conditional hazards to cause-specific cumulative incidence
        proba = torch.cumsum(s_lag * c_haz[..., 1:], dim=1)
        proba = F.pad(proba, pad=(0, 0, 1, 0))

        return interpolate2d(timepoints, self.data_cuts, proba)

    def configure_optimizers(self) -> torch.optim.Optimizer:
        return torch.optim.Adam(self.model.parameters(), self.learning_rate)



import torch
import pytest
import hypothesis
import hypothesis.strategies as st
from einops import repeat

from discotime.utils import KaplanMeier, AalenJohansen, interpolate2d

###############################################################################
# Kaplan-Meier estimator


@pytest.fixture
def km_example():
    """Example data from p. 18 in Kleinbaum & Klein (2005)"""
    data = {
        1: [6, 6, 6, 7, 10, 13, 16, 22, 23],
        0: [6, 9, 10, 11, 17, 19, 20, 25, 32, 32, 34, 45],
    }
    event, time = zip(*[(e, t) for (e, ts) in data.items() for t in ts])
    return KaplanMeier(time, event)


@pytest.mark.parametrize(
    ("time", "expected"),
    [
        # fmt: off
        # single timepoints
        ( 0, 1.0000),
        ( 6, 0.8571),
        (10, 0.7529),
        (22, 0.5378),
        # multiple timepoints
        ([ 6,  7], [0.8571, 0.8067]),
        ([10, 22], [0.7529, 0.5378]),
        # fmt: on
    ],
)
def test_kaplan_meier_estimates(time, expected, km_example):
    assert km_example(time) == pytest.approx(expected, abs=1e-4)


def test_kaplan_meier_percentiles(km_example):
    t = km_example.percentile(km_example._sj)
    assert all(km_example(t) == km_example._sj)


@pytest.mark.parametrize("percentile", [-1.1, 1.1])
def test_kaplan_meier_percentiles_errors(percentile, km_example):
    with pytest.raises(ValueError):
        km_example.percentile(percentile)


# km.percentile should stop giving unique timepoints if p < P(t_max)
def test_kaplan_meier_percentiles_unobserved(km_example):
    t_max, p_min = km_example._tj[-1], km_example._sj[-1]
    assert km_example.percentile(p_min - 0.05) == t_max


###############################################################################
# Aalen-Johansen estimator


@pytest.fixture
def cic(comprisk_testdata):
    time, event = comprisk_testdata
    return AalenJohansen(time, event)


@pytest.mark.parametrize(
    ("timepoint", "cause", "expected"),
    [
        # fmt: off
        # expected obtained with `prodlim` package in R
        ([10.7, 10.8, 10.9], 1, [0.3066, 0.3613, 0.3613]),
        ([ 0.0,  0.6,  0.7], 1, [0.0000, 0.0000, 0.0416]),
        ([ 0.0,  0.6,  0.7], 2, [0.0000, 0.0000, 0.0000]),
        ([17.1, 20.3, 24.4], 1, [0.4488, 0.5362, 0.5362]),
        # fmt: on
    ],
)
def test_cumulative_incidence_comprisk(timepoint, cause, expected, cic):
    estimates = cic(timepoint)[:, cause - 1].squeeze()
    assert estimates == pytest.approx(expected, abs=1e-4)


###############################################################################


@st.composite
def interpolate2d_strategy(draw):
    xp = draw(st.lists(st.integers(0, 20), min_size=2, unique=True))
    x = draw(st.lists(st.integers(min(xp), max(xp)), min_size=1, unique=True))
    coef = draw(st.tuples(st.floats(-10, 10), st.floats(0, 10)))
    return (torch.tensor(sorted(x)), torch.tensor(sorted(xp)), coef)


@hypothesis.given(interpolate2d_strategy())
@hypothesis.example(
    args=(
        torch.tensor([1.0]),
        torch.tensor([0.0000, 0.7500, 1.0000]),
        torch.tensor([1.0, 0.0]),
    )
)
def test_interpolate2d_linear(args):
    x, xp, coef = args

    hypothesis.assume(torch.unique(xp).size() == xp.size())
    hypothesis.assume(torch.unique(x).size() == x.size())

    yp = coef[0] * xp + coef[1]
    yp = repeat(yp, "t -> b t r", b=10, r=2)

    y = coef[0] * x + coef[1]
    y = repeat(y, "t -> b t r", b=10, r=2)

    assert interpolate2d(x, xp, yp) == pytest.approx(y, abs=1e-3)



from typing import Optional, TypeVar
from collections import namedtuple

import numpy as np
import numpy.typing as npt
from numpy.typing import NDArray

from discotime.utils.estimators import AalenJohansen

T_co = TypeVar("T_co", bound=np.generic, covariant=True)

Int = int | np.int_
Num = Int | float | np.float_


def _loo_idx(m: int) -> npt.NDArray[np.int64]:
    """Create indices for leave-one-out resampling.

    Example:
        >>> _loo_idx(4)
        array([[1, 2, 3],
               [0, 2, 3],
               [0, 1, 3],
               [0, 1, 2]])
    """
    return np.arange(1, m, dtype=np.int64) - np.tri(m, m - 1, k=-1, dtype=bool)


def _cic_pv(
    time: npt.NDArray[np.float_],
    event: npt.NDArray[np.int_],
    tau: npt.ArrayLike,
) -> NDArray[np.float64]:
    """Obtain jackknife pseudovalues from the marginal Aalen-Johansen estimate.

    Example:
        >>> rng = np.random.default_rng()
        >>> time, event = rng.weibull(5., size=100), rng.integers(3, size=100)
        >>> aj = _cic_pv(time, event, tau=[0.5, 3, 4])
        >>> aj.shape
        (100, 3, 2)
    """
    N = time.size
    tau = np.asarray(tau, np.float_)

    n_causes = np.max(event)
    cumulative_incidence = AalenJohansen(time, event, n_causes)(tau)
    cumulative_incidence_loo = np.array(
        [AalenJohansen(time[i], event[i], n_causes)(tau) for i in _loo_idx(N)]
    )
    return (N * cumulative_incidence) - ((N - 1) * cumulative_incidence_loo)


class BrierScore:
    """Brier score of survival model with right-censored data.

    In the case of right-censored data, possibly unknown time-dependent event
    status can be replaced with jackknife pseudovalues from the marginal
    Aalen-Johansen estimates [1].

    Args:
        timepoints: sequence of `t` timepoints to evaluate.
        causes: causes for which the brier score is calculated.

    Refs:
        [1] Cortese, Giuliana, Thomas A. Gerds, and Per K. Andersen. "Comparing
        predictions among competing risks models with time‐dependent
        covariates." Statistics in medicine 32.18 (2013): 3089-3101.
    """

    Results = namedtuple("Results", ["null", "model"])

    def __init__(
        self, timepoints: npt.ArrayLike, causes: Optional[list[int]] = None
    ) -> None:
        if causes is not None and 0 in causes:
            raise ValueError("cause 0 is censoring and cannot be included.")
        self.timepoints = np.asanyarray(timepoints)
        self.causes = causes

    @staticmethod
    def _brier_score(label, estimate) -> NDArray[T_co]:
        estimate = np.asanyarray(estimate)
        score = label * (1 - (2 * estimate)) + np.square(estimate)
        return np.atleast_1d(np.mean(score, axis=0))

    def __call__(
        self,
        estimates: Optional[NDArray[T_co]],
        time: npt.ArrayLike,
        event: npt.ArrayLike,
    ) -> Results:
        """Calculate Brier score of survival model.

        Also includes the Brier score of a null-model based on the cumulative
        incidence curve obtained with the Aalen-Johansen estimator.

        Args:
            estimates: an array with shape (`m`, `t`, `e`), where `m` is the
                batch size, `t` is the number of time bins, and `e` is number
                of competing causes/risks.
            time: survival time.
            event: event indicator with event=0 defined censoring.
        """
        time, event = np.asanyarray(time), np.asanyarray(event)

        if not np.issubdtype(event.dtype, np.integer):
            raise TypeError("`event` can only be of integer type.")

        # check dimensionality of input arrays
        if time.ndim != 1:
            raise ValueError(f"`time` is a {time.ndim}D array.")
        if event.ndim != 1:
            raise ValueError(f"`event` is a {event.ndim}D array.")

        # use all available causes if none is specified
        if self.causes is None:
            causes = np.trim_zeros(np.unique(event))
        else:
            causes = np.asanyarray(self.causes)

        # obtain jack-knife pseudovalues
        cic_pv = _cic_pv(time, event, self.timepoints)[..., causes - 1]

        # the null model is the cic from the AalenJohansen estimator
        p_null = AalenJohansen(time, event)(self.timepoints)[..., causes - 1]
        bs_null: npt.NDArray[np.float32] = self._brier_score(cic_pv, p_null)

        # calculate brier score on the model estimates
        if estimates is not None:
            bs_model = self._brier_score(cic_pv, estimates[..., causes - 1])  # type: ignore
        else:
            bs_model = None

        return self.Results(null=bs_null, model=bs_model)



from typing import Optional, Type

import torch
from torch import nn
from torch.nn import functional as F
from einops import rearrange


class Block(nn.Module):
    """Neural network building block for the `Net` class.

    Args:
        n_hidden_units: number of units in each hidden layer.
        add_skip_connection: Defaults to True.
        activation_function: Defaults to
            nn.SiLU.
        batch_normalization: Should batch
            normalization be performed? Defaults to True.
        dropout_rate: dropout_rate is being
            passed along to `nn.Dropout()`. If None, then dropout is not
            being used. Defaults to None.
    """

    def __init__(
        self,
        *,
        n_hidden_units: int,
        add_skip_connection: bool = True,
        activation_function: Type[nn.Module] = nn.SiLU,
        batch_normalization: bool = True,
        dropout_rate: Optional[float] = None,
    ) -> None:
        super().__init__()
        self._should_skip = add_skip_connection
        self.activation_function = activation_function
        self.net = nn.Sequential(
            nn.LazyBatchNorm1d() if batch_normalization else nn.Identity(),
            nn.LazyLinear(out_features=n_hidden_units),
            nn.Dropout(p=dropout_rate) if dropout_rate else nn.Identity(),
            self.activation_function(),
            nn.LazyLinear(out_features=n_hidden_units),
            nn.Dropout(p=dropout_rate) if dropout_rate else nn.Identity(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation_function()(
            x + self.net(x) if self._should_skip else self.net(x)
        )


class Net(nn.Module):
    """
    Feed-forward neural network.

    Has support for dropout, skip connections, and activation function can
    be easily switched out for a different one.

    Args:
        n_out_features : length of output tensor (1D).
        n_blocks (int): Number of `Block()` units included.
        n_hidden_units (int): Number of neurons in each hidden unit.
        add_skip_connection: Defaults to True.
        activation_function: Defaults to
            nn.SiLU.
        batch_normalization: Should batch
            normalization be performed? Defaults to True.
        dropout_rate: dropout_rate is being
            passed along to `nn.Dropout()`. If None, then dropout is not
            being used. Defaults to None.

    :meta private:
    """

    def __init__(
        self,
        n_out_features: int,
        n_blocks: int,
        n_hidden_units: int,
        add_skip_connection: bool = True,
        activation_function: Type[nn.Module] = nn.SiLU,
        batch_normalization: bool = True,
        dropout_rate: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.net = nn.Sequential(
            nn.LazyLinear(out_features=n_hidden_units),
            *(
                Block(
                    n_hidden_units=n_hidden_units,
                    activation_function=activation_function,
                    batch_normalization=batch_normalization,
                    dropout_rate=dropout_rate,
                    add_skip_connection=add_skip_connection,
                )
                for _ in range(n_blocks)
            ),
            nn.LazyLinear(out_features=n_out_features),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)


def negative_log_likelihood(
    logits: torch.Tensor, time: torch.Tensor, event: torch.Tensor
) -> torch.Tensor:
    """Negative log-likelihood for logistic hazard model with competing risks.

    The hazards are expected to be given as logits scale, i.e. they should not
    have been passed through ``torch.log_softmax()`` or similar.

    An implementation of equation (8.6) from Tutz and Schmid [1], inspired by
    the one in ``pycox`` following Kvamme et. al. [2]

    Args:
        logits (:obj:`torch.Tensor`): input logits
        time (:obj:`torch.Tensor`): discretized event times
        event (:obj:`torch.Tensor`): events (0=censored, 1/2/...=events)

    [1]: Tutz, Gerhard, and Matthias Schmid. Modeling discrete time-to-event
    data. New York: Springer, 2016.

    [2]: Kvamme, Håvard, Ørnulf Borgan, and Ida Scheel. "Time-to-event
    prediction with neural networks and Cox regression." arXiv preprint
    arXiv:1907.00825 (2019).
    """

    if logits.ndim != 3:
        raise ValueError(
            "A tensor with exactly three dimensions is expected, "
            f"instead {logits.ndim} dimension was supplied."
        )

    # construct labels that `F.cross_entropy()` can use
    time, event = time.view(-1, 1), event.view(-1, 1)
    target = torch.zeros(logits.shape[:2]).to(time)
    target = target.scatter(dim=1, index=time.long(), src=event)

    return torch.mean(
        (
            F.cross_entropy(
                input=rearrange(logits, "b t r -> b r t"),
                target=target.long(),
                reduction="none",
            )
            .cumsum(dim=1)
            .gather(dim=1, index=time.long())
        )
    )



from pathlib import Path
from typing import Optional
from functools import cached_property

from torch.utils.data import DataLoader, random_split
from sklearn.model_selection import train_test_split
from torch import Generator
import requests
import rdata
import numpy as np
import pandas as pd

import discotime
from discotime.datasets.utils import (
    LitSurvDataModule,
    DataConfig,
    default_fts_transformer,
    SurvDataset,
)
from discotime.datasets.utils import LabelDiscretizer


_ROOT_PATH = Path(discotime.__file__).parent
_DATA_PATH = _ROOT_PATH / "datasets" / "_data" / "_downloads"


class Mgus2(LitSurvDataModule):
    """
    Natural history of 1341 sequential patients with monoclonal gammopathy of
    undetermined significance (MGUS).

    [1]: R. Kyle, T. Therneau, V. Rajkumar, J. Offord, D. Larson, M. Plevak,
        and L. J. Melton III, A long-terms study of prognosis in monoclonal
        gammopathy of undertermined significance. New Engl J Med, 346:564-569
        (2002).
    """

    n_features = 7
    n_risks = 2

    def __init__(
        self,
        data_config: DataConfig,
        data_dir: Path = _DATA_PATH,
        seed: int = 13411341,
    ) -> None:
        super().__init__()

        self.seed = seed
        self.rng = Generator().manual_seed(seed)
        self.mgus2_csv = data_dir / "mgus2.csv"

        self.fts_transformer = default_fts_transformer()
        self.lab_transformer = LabelDiscretizer(
            n_bins=data_config.n_time_bins,
            scheme=data_config.discretization_scheme,
        )

        self._config = data_config
        self._dset_fit: Optional[SurvDataset] = None
        self._dset_test: Optional[SurvDataset] = None

        self.save_hyperparameters(ignore=["data_dir"])

    def prepare_data(self) -> None:
        """Fetch the built-in mgus2 from the R survival package.

        The mgus2 data [1] is extracted from the rdata file and saved as a csv
        in the discotime installation directory. If the csv is already
        available, then the download logic is skipped.

        [1]: Therneau T (2023). A Package for Survival Analysis in R.
        """
        url = "https://github.com/therneau/survival/raw/649851/data/cancer.rda"

        if not self.mgus2_csv.is_file():
            # download and parse rdata file
            r = requests.get(url)
            parsed = rdata.parser.parse_data(r.content)
            df = rdata.conversion.convert(parsed)["mgus2"]
            # slight reformatting
            x = df.loc[:, "age":"mspike"]  # type: ignore
            t = df[["futime", "ptime"]].min(axis=1)
            e = pd.concat((df["pstat"] * 2, df["death"]), axis=1).max(axis=1)
            y = pd.concat({"time": t, "event": e}, axis=1)
            # combine and save as csv
            pd.concat((x, y), axis=1).to_csv(self.mgus2_csv, index=False)

    @property
    def dset_fit(self) -> SurvDataset:
        if self._dset_fit is None:
            raise AttributeError(
                "`dset_fit` not initialized yet."
                "Maybe try calling `.setup()` first?"
            )
        return self._dset_fit

    @property
    def dset_test(self) -> SurvDataset:
        if self._dset_test is None:
            raise AttributeError(
                "`dset_fit` not initialized yet."
                "Maybe try calling `.setup()` first?"
            )
        return self._dset_test

    @cached_property
    def time_range(self):
        data = pd.read_csv(self.mgus2_csv, dtype={"event": np.int32})
        data = data[data.event != 0]
        time = data.groupby(["event"])["time"]
        return (time.min().max(), time.max().min())

    def setup(self, stage: Optional[str] = None) -> None:
        if self._dset_fit is None:
            data = pd.read_csv(self.mgus2_csv, dtype={"event": np.int32})
            features = data.loc[:, "age":"mspike"]  # type: ignore
            outcomes = data.loc[:, ["time", "event"]]

            X_fit, X_test, Y_fit, Y_test = train_test_split(
                features, outcomes, test_size=0.2, random_state=self.seed
            )

            # prepare transformers
            self.lab_transformer.fit(Y_fit.time.values, Y_fit.event.values)
            self.fts_transformer.fit(X_fit)

            # initialize the datasets
            self._dset_fit = SurvDataset(
                features=self.fts_transformer.transform(X_fit),
                event_time=Y_fit.time.values,
                event_status=Y_fit.event.values,
                discretizer=self.lab_transformer,
            )
            self._dset_test = SurvDataset(
                features=self.fts_transformer.transform(X_test),
                event_time=Y_test.time.values,
                event_status=Y_test.event.values,
                discretizer=self.lab_transformer,
            )

        if stage in {"fit", "validate", "debug"}:
            self.dset_train, self.dset_val = random_split(
                dataset=self.dset_fit,
                lengths=[0.8, 0.2],
                generator=self.rng,
            )

    def train_dataloader(self):
        return DataLoader(
            self.dset_train,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=1,
        )

    def val_dataloader(self):
        return DataLoader(
            self.dset_val, batch_size=self.batch_size, num_workers=1
        )

    def test_dataloader(self):
        return DataLoader(
            self.dset_test, batch_size=self.batch_size, num_workers=1
        )

    def predict_dataloader(self):
        return DataLoader(
            self.dset_test, batch_size=self.batch_size, num_workers=1
        )



import pytest
from lightning.pytorch import Trainer
from warnings import filterwarnings
from pathlib import Path

from discotime.models import ModelConfig, LitSurvModule
from discotime.datasets import DataConfig, Mgus2


@pytest.fixture
def lit_data_module():
    return Mgus2(DataConfig())


@pytest.fixture
def lit_module():
    return LitSurvModule(ModelConfig(n_sequential_blocks=10))


def test_load_checkpoint_works(
    lit_module: LitSurvModule, lit_data_module: Mgus2, tmp_path: Path
):
    filterwarnings("ignore", message=".*num_workers")
    filterwarnings("ignore", message=".*self.log")

    trainer = Trainer(barebones=True, max_epochs=1)
    trainer.fit(lit_module, lit_data_module)
    trainer.save_checkpoint(tmp_path / "checkpoint.ckpt")

    # this needs to work without any errors
    lit_module = lit_module.__class__.load_from_checkpoint(
        checkpoint_path=tmp_path / "checkpoint.ckpt", strict=True
    )



from typing import Optional
from collections.abc import Iterable
from dataclasses import dataclass
from abc import ABCMeta, abstractmethod
import warnings

import numpy as np
import numpy.typing as npt

from torch.utils.data import Dataset
import lightning.pytorch as pl

from sklearn.compose import make_column_selector, make_column_transformer
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer

from discotime.utils import KaplanMeier
from discotime.utils.typing import Int, Num, LabelTransformer, SurvData

###############################################################################


class SurvDataset(Dataset):
    r"""Assemble a survival dataset for discrete-time survival analysis.

    A discrete time survival dataset :math:`\mathfrak{D}` is a set of :math:`n`
    tuples :math:`(t_{i}, \delta_{i}, \mathbf{x}_{i})` where :math:`(t_i = \min
    \{T_i, C_i\})` is the event time, :math:`\delta_{i} \in \{0, ..., m\}`
    is the event indicator (with :math:`(\delta_i = 0)` defined as censoring),
    and :math:`\mathbf{x}_{i} \in \mathbb{R}^d` is a :math:`d`-dimensional
    vector of time-independent predictors or covariates.

    Args:
        features: time-independent features.
        event_time: follow-up time (continuous).
        event_status: event indicator (0=censored, 1/2/...=competing risks).
        discretizer: discretizer that follows the :obj:`LabelTransformer`
            protocol that convert continuous time/event tuples to their
            respective discretized versions. Typically this would be
            :obj:`LabelDiscretizer` unless a custom discretization object is
            used.
    """

    def __init__(
        self,
        features: npt.ArrayLike,
        event_time: npt.ArrayLike,
        event_status: npt.ArrayLike,
        discretizer: LabelTransformer,
    ) -> None:
        super().__init__()

        ct = np.asanyarray(event_time, dtype=np.int32)
        ce = np.asanyarray(event_status, dtype=np.int32)
        dt, de = discretizer.transform(ct, ce)

        self.features = np.asanyarray(features, dtype=np.float32)
        if not self.features.shape[0] == len(ct) == len(ce):
            raise ValueError(
                "`features`, `event_time`, and `event_status`"
                "should all have the same number of rows."
            )

        self.event_time_cont = ct
        self.event_status_cont = ce
        self.event_time_disc = dt
        self.event_status_disc = de

    def __len__(self):
        return len(self.event_time_cont)

    def __getitem__(self, index) -> SurvData:
        return SurvData(
            features=self.features[index, ...],
            event_time_disc=self.event_time_disc[index],
            event_status_disc=self.event_status_disc[index],
            event_time_cont=self.event_time_cont[index],
            event_status_cont=self.event_status_cont[index],
        )


###############################################################################


@dataclass(kw_only=True)
class DataConfig:
    """Configuration class for data modules."""

    batch_size: int = 32
    """
    The batch size defines the number of samples that will be propagated
    through the network at each training step.
    """
    n_time_bins: int = 20
    """
    Specifies the size of the discretization grid.
    A default of around 20-30 usually works good.
    """
    discretization_scheme: str = "number"
    discretization_grid: Optional[list[float]] = None
    max_time: Optional[float] = None


class LitSurvDataModule(pl.LightningDataModule, metaclass=ABCMeta):
    def __init__(self) -> None:
        super().__init__()
        self._dset_fit: Optional[SurvDataset] = None
        self._dset_test: Optional[SurvDataset] = None
        self._config: Optional[DataConfig] = None

    def __repr__(self) -> str:
        return (
            super().__repr__() + "\n    "
            "n_features: {}, n_risks: {}, n_time_bins: {}".format(
                self.n_features, self.n_risks, self.n_time_bins
            )
        )

    @property
    def config(self) -> DataConfig:
        if self._config is None:
            raise AttributeError("`self._config` needs to be defined.")
        return self._config

    @property
    def cuts(self) -> npt.NDArray[np.float_]:
        return self.lab_transformer.cuts

    @property
    def n_time_bins(self) -> int:
        return self.config.n_time_bins

    @property
    def batch_size(self) -> int:
        return self.config.batch_size

    @property
    @abstractmethod
    def time_range(self) -> tuple[Num, Num]:
        ...

    @property
    @abstractmethod
    def n_features(self) -> int:
        ...

    @property
    @abstractmethod
    def n_risks(self) -> int:
        ...

    @abstractmethod
    def setup(self, stage: Optional[str] = None) -> None:
        ...

    @property
    def lab_transformer(self) -> LabelTransformer:
        return self._lab_transformer

    @lab_transformer.setter
    def lab_transformer(self, value: LabelTransformer) -> None:
        self._lab_transformer = value


###############################################################################


def default_fts_transformer():
    transformer = make_column_transformer(
        # apply one-hot encoding to str columns
        (
            OneHotEncoder(handle_unknown="ignore"),
            make_column_selector(dtype_include=object),
        ),
        # apply z-scaling and mean-imputation to numeric colums
        (
            Pipeline(
                [
                    ("scaler", StandardScaler()),
                    ("impute", SimpleImputer(strategy="constant")),
                ]
            ),
            make_column_selector(dtype_include=np.number),
        ),
        remainder="passthrough",
    )
    return transformer


###############################################################################


def _discretize_labels(
    time: Iterable[Num],
    event: Iterable[Int],
    cuts: Iterable[Num],
) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
    """Discretize continous time/event pairs using supplied cuts."""
    _time, _event, _cuts = map(np.asanyarray, (time, event, cuts))

    maxi = len(_cuts) - 1
    de = np.atleast_1d(np.copy(_event)).astype(np.int64)
    de[_time > _cuts[-1]] = 0  # censored

    _time = np.atleast_1d(np.clip(_time, a_min=0, a_max=_cuts[-1]))
    dt = (np.searchsorted(_cuts, _time, side="right") - 1).astype(np.int64)

    # give credit for surviving more than half of an interval
    where = (de == 0) & (dt != maxi)
    x1, x2 = _cuts[dt[where]], _cuts[dt[where] + 1]
    dt[where] += (_time[where] - x1) > (x2 - _time[where])

    dt[dt == maxi] -= 1
    return (dt, de)


def _cuts_interval(n_bins: Int, max_time: Num) -> npt.NDArray[np.float64]:
    """Create cuts that define n groups with equal range.

    Example:
        >>> _cuts_interval(5, 10)
        array([ 0.,  2.,  4.,  6.,  8., 10.])
    """

    return np.linspace(0, max_time, n_bins + 1, dtype=np.float64)


def _cuts_number(
    n_bins: Int,
    max_time: Num,
    time: Iterable[Num],
    event: Iterable[Int],
) -> npt.NDArray[np.float64]:
    """Makes n groups with approx. equal number of observations."""
    _time, _event = map(np.asarray, (time, event))
    km = KaplanMeier(_time, (_event != 0).astype(np.int_))
    qs = np.linspace(1, km(max_time).item(), n_bins + 1)
    return km.percentile(qs, dtype=np.float64)


class LabelDiscretizer:
    """Discretize continous time/event pairs.

    The class can either learn a discretization grid from the training data
    using one of the built-in discretization schemes, or the user can supply an
    iterable with cut points.

    Implementation heavily inspired by pycox.preprocessing.label_tranform [1].

    [1]: Kvamme, Håvard, Ørnulf Borgan, and Ida Scheel. "Time-to-event
    prediction with neural networks and Cox regression." arXiv preprint
    arXiv:1907.00825 (2019).
    """

    def __init__(
        self,
        scheme: Optional[str] = None,
        n_bins: Optional[int] = None,
        *,
        cut_points: Optional[Iterable[Num]] = None,
        max_time: Optional[Num] = None,
    ) -> None:
        """Initialize LabTransDiscrete

        Available discretization schemes are:

        * interval: makes n groups with equal range
        * number: makes n groups with approx. equal number of observations
        * manual: supplying an iterable of cutpoints triggers the class to use
            those numbers to define the intervals instead.

        Args:
            n_bins: number of discretization bins to obtain.
            scheme: discretization scheme used to obtain cuts.
            cut_points: an iterable of pre-calculated cuts. Should only be
                used for manual specification of cut points. Default is [].
            max_time: truncate follow-up time. Default is None.
        """

        self._cuts = np.array([], np.float64)
        self._max_time = max_time

        if scheme in {"interval", "number"}:
            if isinstance(n_bins, int):
                self.n_bins = n_bins
                self.scheme = scheme
            else:
                raise ValueError("`n_bins` is not specified.")
        elif cut_points is not None:
            self._cuts = np.asarray(cut_points, np.float64)
            self.n_bins = len(self._cuts) - 1
            self.scheme = "manual"
        else:
            raise Exception(f"{self.__class__} could not be instantiated.")

    @property
    def max_time(self) -> Num:
        if self._max_time is None:
            raise AttributeError(
                "`max_time` not defined yet. Try calling `.fit()` first?"
            )
        return self._max_time

    @property
    def cuts(self) -> npt.NDArray[np.float64]:
        if not self._cuts.size:
            raise AttributeError(
                "`cuts` not defined yet. Try calling `.fit()` first?"
            )
        return self._cuts

    def fit(self, time: Iterable[Num], event: Iterable[Int]) -> None:
        self._max_time = self._max_time or max(time)

        match self.scheme:
            case "interval":
                self._cuts = _cuts_interval(self.n_bins, self.max_time)
            case "number":
                self._cuts = _cuts_number(
                    self.n_bins, self.max_time, time, event
                )
            case "manual":
                warnings.warn(
                    "ignoring fit(), scheme is set to 'manual'", UserWarning
                )

        if not np.unique(self.cuts).size == (self.n_bins + 1):
            raise ValueError(
                "Not all values in `self.cuts` are unique."
                f"Decrease the size of `n_bins` (currently {self.n_bins})"
            )

    def fit_transform(
        self, time: Iterable[Num], event: Iterable[Int]
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        self.fit(time, event)
        return self.transform(time, event)

    def transform(
        self, time: Iterable[Num], event: Iterable[Int]
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        return _discretize_labels(time, event, self.cuts)



from typing import Optional
import numpy as np
import numpy.typing as npt
from discotime.metrics import BrierScore


class IPA:
    """Index of prediction accuracy (IPA) for right-censored survival models.

    In the context of survival analysis, the IPA is also known as the Brier
    skill score (BSS).
    """

    def __init__(
        self,
        timepoints: npt.NDArray[np.float_],
        causes: Optional[list[int]] = None,
        integrate: bool = True,
    ) -> None:
        """
        Args:
            timepoints: sequence of `t` timepoints to evaluate.
            causes: causes for which the brier score is calculated.
            integrate: should the time-dependent IPA be integrated?
                (default: True)
        """
        self.timepoints = np.asanyarray(timepoints)
        self.brier_score = BrierScore(self.timepoints, causes)
        self.integrate = integrate

    def __call__(
        self,
        estimates: npt.NDArray[np.floating],
        time: npt.NDArray[np.floating],
        event: npt.NDArray[np.integer],
    ) -> np.float_:
        """Calculate the index of prediction accuracy.

        Args:
            estimates: array[batch, timepoints, cause]
                model estimates of cause probabilities.
            time: observed follow-up.
            event: event indicator at time with 0 as censoring.

        If there are multiple timepoints to consider, then the IPA is
        numerically integrated using the trapezoidal method.
        """

        bs = self.brier_score(estimates, time, event)

        with np.errstate(divide="ignore"):
            ipa = 1 - (bs.model / bs.null)
            ipa = np.nan_to_num(ipa, neginf=0, posinf=0)

        if len(self.timepoints) > 1 and self.integrate:
            scaling = 1 / (max(self.timepoints) - min(self.timepoints))
            ipa = np.trapz(ipa, self.timepoints, axis=0) * scaling

        return ipa



import pytest
from pytest import approx
from warnings import filterwarnings
import torch
import torch.nn.functional as F

from discotime.models import ModelConfig, LitSurvModule
from discotime.datasets import Mgus2, DataConfig
from discotime.models.components import negative_log_likelihood


def test_nll_loss():
    time = torch.tensor([0, 0, 1, 2, 3])
    event = torch.tensor([1, 2, 0, 0, 1])
    logits = torch.rand((5, 4, 3))
    phi = F.log_softmax(logits, dim=2)

    # calculate loss "manually" from the tutz formula
    loss0 = 0
    for i, (_time, _event) in enumerate(zip(time, event)):
        for t in range(_time + 1):
            if t == _time:
                loss0 -= phi[i][t][_event]
            else:
                loss0 -= phi[i][t][0]  # no event
    loss0 = loss0 / len(time)
    loss1 = negative_log_likelihood(logits, time, event)

    assert loss1 == pytest.approx(loss0, abs=1e-6)


def test_epoch_loss_batch_averaging():
    """Total epoch loss should be independent of batch size."""

    filterwarnings("ignore", message=".*`self.log")

    mgus_bs60 = Mgus2(DataConfig(batch_size=60))
    mgus_bs30 = Mgus2(DataConfig(batch_size=30))
    mgus_bs30.prepare_data()

    model = LitSurvModule(
        ModelConfig(
            n_risks=mgus_bs30.n_risks,
            n_time_bins=mgus_bs30.n_time_bins,
        )
    )
    model.datamodule = mgus_bs30
    model.setup(stage="fit")

    def get_epoch_loss(dataset):
        dataset.setup(stage="fit")
        with torch.set_grad_enabled(False):
            model.eval()
            losses = (
                model.validation_step(batch, batch_idx) * len(batch[1])
                for batch_idx, batch in enumerate(dataset.train_dataloader())
            )
            return sum(losses) / len(dataset.dset_train)

    assert get_epoch_loss(mgus_bs30) == approx(get_epoch_loss(mgus_bs60))
